from typing import Union
import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, slice_candles


@njit(cache=True)
def _bb_width(source: np.ndarray, period: int, mult: float) -> np.ndarray:
    """
    Calculate Bollinger Bands Width using Numba
    """
    n = len(source)
    bbw = np.full(n, np.nan)
    
    if n < period:
        return bbw
        
    # Pre-calculate sum and sum of squares for optimization
    sum_x = np.zeros(n - period + 1)
    sum_x2 = np.zeros(n - period + 1)
    
    # Initial window
    sum_x[0] = np.sum(source[:period])
    sum_x2[0] = np.sum(source[:period] ** 2)
    
    # Rolling window calculations
    for i in range(1, n - period + 1):
        sum_x[i] = sum_x[i-1] - source[i-1] + source[i+period-1]
        sum_x2[i] = sum_x2[i-1] - source[i-1]**2 + source[i+period-1]**2
    
    # Calculate mean and standard deviation
    mean = sum_x / period
    std = np.sqrt((sum_x2 / period) - (mean ** 2))
    
    # Calculate BBW
    for i in range(period - 1, n):
        idx = i - period + 1
        basis = mean[idx]
        upper = basis + mult * std[idx]
        lower = basis - mult * std[idx]
        bbw[i] = (upper - lower) / basis
        
    return bbw


def bollinger_bands_width(candles: np.ndarray, period: int = 20, mult: float = 2.0, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    BBW - Bollinger Bands Width - Bollinger Bands Bandwidth using Numba for optimization

    :param candles: np.ndarray
    :param period: int - default: 20
    :param mult: float - default: 2
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)
    
    result = _bb_width(source, period, mult)
    
    return result if sequential else result[-1]



from typing import Union

import numpy as np
from numba import njit
from jesse.helpers import slice_candles


@njit(cache=True)
def compute_multiplier(high: np.ndarray, low: np.ndarray, close: np.ndarray) -> np.ndarray:
    n = high.shape[0]
    out = np.empty(n, dtype=high.dtype)
    for i in range(n):
        rng = high[i] - low[i]
        if rng != 0:
            out[i] = ((close[i] - low[i]) - (high[i] - close[i])) / rng
        else:
            out[i] = 0.0
    return out


@njit(cache=True)
def elementwise_multiply(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    n = a.shape[0]
    out = np.empty(n, dtype=a.dtype)
    for i in range(n):
        out[i] = a[i] * b[i]
    return out


@njit(cache=True)
def cumulative_sum(arr: np.ndarray) -> np.ndarray:
    n = arr.shape[0]
    out = np.empty(n, dtype=arr.dtype)
    total = 0.0
    for i in range(n):
        total += arr[i]
        out[i] = total
    return out


@njit(cache=True)
def ema(values: np.ndarray, period: int) -> np.ndarray:
    n = values.shape[0]
    result = np.empty(n, dtype=values.dtype)
    alpha = 2.0 / (period + 1)
    result[0] = values[0]
    for i in range(1, n):
        result[i] = alpha * values[i] + (1.0 - alpha) * result[i - 1]
    return result


@njit(cache=True)
def subtract_arrays(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    n = a.shape[0]
    out = np.empty(n, dtype=a.dtype)
    for i in range(n):
        out[i] = a[i] - b[i]
    return out


def adosc(candles: np.ndarray, fast_period: int = 3, slow_period: int = 10, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    ADOSC - Chaikin A/D Oscillator (Numba accelerated version)

    :param candles: np.ndarray of candles
    :param fast_period: int - default: 3
    :param slow_period: int - default: 10
    :param sequential: bool - default: False
    :return: float or np.ndarray
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    volume = candles[:, 5]

    multiplier = compute_multiplier(high, low, close)
    mf_volume = elementwise_multiply(multiplier, volume)
    ad_line = cumulative_sum(mf_volume)

    fast_ema = ema(ad_line, fast_period)
    slow_ema = ema(ad_line, slow_period)
    adosc_vals = subtract_arrays(fast_ema, slow_ema)

    return adosc_vals if sequential else adosc_vals[-1]



from collections import namedtuple
import numpy as np
from numba import jit
from jesse.helpers import get_candle_source, slice_candles

Wavetrend = namedtuple('Wavetrend', ['wt1', 'wt2', 'wtCrossUp', 'wtCrossDown', 'wtOversold', 'wtOverbought', 'wtVwap'])


# Wavetrend indicator ported from:  https://www.tradingview.com/script/Msm4SjwI-VuManChu-Cipher-B-Divergences/
#                                   https://www.tradingview.com/script/2KE8wTuF-Indicator-WaveTrend-Oscillator-WT/
#
# buySignal = wtCross and wtCrossUp and wtOversold
# sellSignal = wtCross and wtCrossDown and wtOverbought
#
# See https://github.com/ysdede/lazarus3/blob/partialexit/strategies/lazarus3/__init__.py for working jesse.ai example.

@jit(cache=True)
def _ema(arr: np.ndarray, period: int) -> np.ndarray:
    """
    Calculate EMA using Numba-optimized implementation
    """
    alpha = 2.0 / (period + 1.0)
    result = np.zeros_like(arr)
    result[0] = arr[0]
    
    for i in range(1, len(arr)):
        result[i] = alpha * arr[i] + (1 - alpha) * result[i-1]
    
    return result

@jit(cache=True)
def _sma(arr: np.ndarray, period: int) -> np.ndarray:
    """
    Calculate SMA using Numba-optimized implementation
    """
    result = np.zeros_like(arr)
    
    # Handle first period-1 elements
    cumsum = 0.0
    for i in range(period-1):
        cumsum += arr[i]
        result[i] = cumsum / (i + 1)
    
    # Handle remaining elements with sliding window
    cumsum = sum(arr[0:period])
    result[period-1] = cumsum / period
    
    for i in range(period, len(arr)):
        cumsum = cumsum - arr[i-period] + arr[i]
        result[i] = cumsum / period
    
    return result

@jit(cache=True)
def _fast_wt(src: np.ndarray, wtchannellen: int, wtaveragelen: int, wtmalen: int) -> tuple:
    """
    Calculate Wavetrend components using Numba
    """
    esa = _ema(src, wtchannellen)
    
    # Calculate absolute difference
    abs_diff = np.abs(src - esa)
    de = _ema(abs_diff, wtchannellen)
    
    # Avoid division by zero
    de = np.where(de == 0, 1e-10, de)
    ci = (src - esa) / (0.015 * de)
    
    wt1 = _ema(ci, wtaveragelen)
    wt2 = _sma(wt1, wtmalen)
    
    return wt1, wt2

def wt(candles: np.ndarray, wtchannellen: int = 9, wtaveragelen: int = 12, wtmalen: int = 3, oblevel: int = 53, oslevel: int = -53, source_type: str = "hlc3", sequential: bool = False) -> Wavetrend:
    """
    Wavetrend indicator

    :param candles: np.ndarray
    :param wtchannellen:  int - default: 9
    :param wtaveragelen: int - default: 12
    :param wtmalen: int - default: 3
    :param oblevel: int - default: 53
    :param oslevel: int - default: -53
    :param source_type: str - default: "hlc3"
    :param sequential: bool - default: False

    :return: Wavetrend
    """
    candles = slice_candles(candles, sequential)
    src = get_candle_source(candles, source_type=source_type)
    
    # Convert inputs to float64 for better numerical stability
    src = src.astype(np.float64)
    
    # Calculate main components
    wt1, wt2 = _fast_wt(src, wtchannellen, wtaveragelen, wtmalen)
    
    # Calculate additional components
    wtVwap = wt1 - wt2
    wtOversold = wt2 <= oslevel
    wtOverbought = wt2 >= oblevel
    wtCrossUp = wt2 - wt1 <= 0
    wtCrossDown = wt2 - wt1 >= 0

    if sequential:
        return Wavetrend(wt1, wt2, wtCrossUp, wtCrossDown, wtOversold, wtOverbought, wtVwap)
    else:
        return Wavetrend(wt1[-1], wt2[-1], wtCrossUp[-1], wtCrossDown[-1], wtOversold[-1], wtOverbought[-1], wtVwap[-1])



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, same_length, slice_candles


def vwma(candles: np.ndarray, period: int = 20, source_type: str = "close", sequential: bool = False) -> Union[
        float, np.ndarray]:
    """
    VWMA - Volume Weighted Moving Average

    :param candles: np.ndarray
    :param period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
        volume = np.ones_like(candles)
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)
        volume = candles[:, 5]

    # Calculate price * volume
    weighted = source * volume
    
    # Compute cumulative sums for weighted price and volume
    cumsum_weighted = np.cumsum(weighted)
    cumsum_volume = np.cumsum(volume)

    # Allocate array for VWMA values
    vwma_vals = np.empty_like(source, dtype=float)

    # For initial indices where full period is not available, use cumulative sums
    vwma_vals[:period] = cumsum_weighted[:period] / np.where(cumsum_volume[:period] == 0, 1, cumsum_volume[:period])

    # For indices with a complete period, use the rolling difference of cumulative sums
    vwma_vals[period:] = (cumsum_weighted[period:] - cumsum_weighted[:-period]) / np.where((cumsum_volume[period:] - cumsum_volume[:-period]) == 0, 1, (cumsum_volume[period:] - cumsum_volume[:-period]))

    return same_length(candles, vwma_vals) if sequential else vwma_vals[-1]



from collections import namedtuple
import numpy as np
from numba import njit

IchimokuCloud = namedtuple('IchimokuCloud', ['conversion_line', 'base_line', 'span_a', 'span_b'])


@njit
def _get_period_hl(candles: np.ndarray, period: int) -> tuple:
    """
    Calculate period high and low using Numba
    """
    period_high = candles[-period:, 3].max()  # high prices
    period_low = candles[-period:, 4].min()   # low prices
    return period_high, period_low


@njit
def _calculate_ichimoku(candles: np.ndarray, conversion_line_period: int, 
                       base_line_period: int, lagging_line_period: int,
                       displacement: int) -> tuple:
    """
    Core Ichimoku Cloud calculation using Numba
    """
    # Calculate for earlier period (displaced)
    earlier_candles = candles[:-(displacement - 1)]
    
    # Earlier periods calculations
    small_ph, small_pl = _get_period_hl(earlier_candles, conversion_line_period)
    mid_ph, mid_pl = _get_period_hl(earlier_candles, base_line_period)
    long_ph, long_pl = _get_period_hl(earlier_candles, lagging_line_period)
    
    early_conversion_line = (small_ph + small_pl) / 2
    early_base_line = (mid_ph + mid_pl) / 2
    span_a = (early_conversion_line + early_base_line) / 2
    span_b = (long_ph + long_pl) / 2
    
    # Current period calculations
    current_small_ph, current_small_pl = _get_period_hl(candles, conversion_line_period)
    current_mid_ph, current_mid_pl = _get_period_hl(candles, base_line_period)
    
    current_conversion_line = (current_small_ph + current_small_pl) / 2
    current_base_line = (current_mid_ph + current_mid_pl) / 2
    
    return current_conversion_line, current_base_line, span_a, span_b


def ichimoku_cloud(candles: np.ndarray, conversion_line_period: int = 9, base_line_period: int = 26,
                   lagging_line_period: int = 52, displacement: int = 26) -> IchimokuCloud:
    """
    Ichimoku Cloud

    :param candles: np.ndarray
    :param conversion_line_period: int - default: 9
    :param base_line_period: int - default: 26
    :param lagging_line_period: int - default: 52
    :param displacement: - default: 26

    :return: IchimokuCloud(conversion_line, base_line, span_a, span_b)
    """
    if candles.shape[0] < 80:
        return IchimokuCloud(np.nan, np.nan, np.nan, np.nan)

    if candles.shape[0] > 80:
        candles = candles[-80:]
        
    conversion_line, base_line, span_a, span_b = _calculate_ichimoku(
        candles, conversion_line_period, base_line_period, 
        lagging_line_period, displacement
    )
    
    return IchimokuCloud(conversion_line, base_line, span_a, span_b)



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

DonchianChannel = namedtuple('DonchianChannel', ['upperband', 'middleband', 'lowerband'])


def donchian(candles: np.ndarray, period: int = 20, sequential: bool = False) -> DonchianChannel:
    """
    Donchian Channels

    :param candles: np.ndarray
    :param period: int - default: 20
    :param sequential: bool - default: False

    :return: DonchianChannel(upperband, middleband, lowerband)
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]

    if sequential:
        # Compute rolling maximum and minimum using sliding_window_view, vectorized without explicit loops
        from numpy.lib.stride_tricks import sliding_window_view
        n = high.shape[0]
        # Prepare output arrays with NaN for the initial period-1 values
        rolling_max = np.empty(n)
        rolling_min = np.empty(n)
        rolling_max[:period - 1] = np.nan
        rolling_min[:period - 1] = np.nan
        # Compute sliding window view for the valid windows
        windowed_high = sliding_window_view(high, window_shape=period)
        windowed_low = sliding_window_view(low, window_shape=period)
        rolling_max[period - 1:] = np.max(windowed_high, axis=1)
        rolling_min[period - 1:] = np.min(windowed_low, axis=1)
        middleband = (rolling_max + rolling_min) / 2
        return DonchianChannel(rolling_max, middleband, rolling_min)
    else:
        # Non-sequential: compute only the last period's max and min
        uc = np.max(high[-period:])
        lc = np.min(low[-period:])
        mc = (uc + lc) / 2
        return DonchianChannel(uc, mc, lc)



from typing import Union

import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, slice_candles

@njit(cache=True)
def _ema(source: np.ndarray, period: int) -> np.ndarray:
    alpha = 2.0 / (period + 1.0)
    result = np.zeros_like(source)
    result[0] = source[0]
    
    for i in range(1, len(source)):
        result[i] = alpha * source[i] + (1 - alpha) * result[i-1]
        
    return result

def tema(candles: np.ndarray, period: int = 9, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    TEMA - Triple Exponential Moving Average

    :param candles: np.ndarray
    :param period: int - default: 9
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    ema1 = _ema(source, period)
    ema2 = _ema(ema1, period)
    ema3 = _ema(ema2, period)
    res = 3 * ema1 - 3 * ema2 + ema3
    
    return res if sequential else res[-1]




from collections import namedtuple
import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, np_shift, slice_candles

AG = namedtuple('AG', ['jaw', 'teeth', 'lips'])

def smma(source: np.ndarray, length: int) -> np.ndarray:
    return _smma_numba(source, length)

@njit
def _smma_numba(source, length):
    alpha = 1.0 / length
    total = 0.0
    for i in range(length):
        total += source[i]
    init_val = total / length
    N = len(source)
    result = np.empty(N, dtype=np.float64)
    result[0] = alpha * source[0] + (init_val * (1 - alpha))
    for i in range(1, N):
        result[i] = alpha * source[i] + (1 - alpha) * result[i-1]
    return result

def alligator(candles: np.ndarray, source_type: str = "hl2", sequential: bool = False) -> AG:
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    jaw = np_shift(smma(source, 13), 8, fill_value=np.nan)
    teeth = np_shift(smma(source, 8), 5, fill_value=np.nan)
    lips = np_shift(smma(source, 5), 3, fill_value=np.nan)

    if sequential:
        return AG(jaw, teeth, lips)
    else:
        return AG(jaw[-1], teeth[-1], lips[-1])



from starlette.responses import JSONResponse

from jesse.modes.import_candles_mode import CandleExchange
from jesse.modes.import_candles_mode.drivers import drivers, driver_names
from jesse.services.redis import sync_redis


def get_exchange_supported_symbols(exchange: str) -> JSONResponse:
    # first try to get from cache
    cache_key = f'exchange-symbols:{exchange}'
    cached_result = sync_redis.get(cache_key)
    if cached_result is not None:
        return JSONResponse({
            'data': eval(cached_result)
        }, status_code=200)

    arr = []

    try:
        driver: CandleExchange = drivers[exchange]()
    except KeyError:
        raise ValueError(f'{exchange} is not a supported exchange. Supported exchanges are: {driver_names}')

    try:
        arr = driver.get_available_symbols()
        # cache successful result for 5 minutes
        sync_redis.setex(cache_key, 300, str(arr))
    except Exception as e:
        return JSONResponse({
            'error': str(e)
        }, status_code=500)

    return JSONResponse({
        'data': arr
    }, status_code=200)



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import slice_candles


def chop(candles: np.ndarray, period: int = 14, scalar: float = 100, drift: int = 1, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Choppiness Index (CHOP)

    :param candles: np.ndarray
    :param period: int - default: 14
    :param scalar: float - default: 100
    :param drift: int - default: 1
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    # Preprocess candles using original slicing
    candles = slice_candles(candles, sequential)
    res = _chop_numba(candles, period, scalar, drift)
    return res if sequential else res[-1]


@njit(cache=True)
def _chop_numba(candles, period, scalar, drift):
    n = candles.shape[0]
    # Extract the necessary candle values
    close = candles[:, 2]
    high = candles[:, 3]
    low = candles[:, 4]

    # Compute True Range (TR)
    tr = np.empty(n, dtype=np.float64)
    tr[0] = high[0] - low[0]
    for i in range(1, n):
        d1 = high[i] - low[i]
        d2 = abs(high[i] - close[i - 1])
        d3 = abs(low[i] - close[i - 1])
        if d1 >= d2 and d1 >= d3:
            tr[i] = d1
        elif d2 >= d1 and d2 >= d3:
            tr[i] = d2
        else:
            tr[i] = d3

    # Compute Average True Range (ATR)
    atr = np.empty(n, dtype=np.float64)
    if drift == 1:
        for i in range(n):
            atr[i] = tr[i]
    else:
        # Set initial values to NaN for indices before drift - 1
        for i in range(drift - 1):
            atr[i] = np.nan
        # Compute initial ATR as the average of the first 'drift' TR values
        total = 0.0
        for i in range(drift):
            total += tr[i]
        initial_atr = total / drift
        atr[drift - 1] = initial_atr
        for i in range(drift, n):
            atr[i] = (atr[i - 1] * (drift - 1) + tr[i]) / drift

    # Compute rolling sum of ATR over 'period'
    atr_sum = np.empty(n, dtype=np.float64)
    for i in range(n):
        atr_sum[i] = np.nan
    for i in range(period - 1, n):
        valid = True
        window_sum = 0.0
        for j in range(i - period + 1, i + 1):
            if np.isnan(atr[j]):
                valid = False
                break
            window_sum += atr[j]
        if valid:
            atr_sum[i] = window_sum
        else:
            atr_sum[i] = np.nan

    # Compute rolling maximum of high over 'period'
    hh = np.empty(n, dtype=np.float64)
    for i in range(n):
        hh[i] = np.nan
    for i in range(period - 1, n):
        max_val = -1e100  # very low number as initial max
        valid = True
        for j in range(i - period + 1, i + 1):
            if np.isnan(high[j]):
                valid = False
                break
            if high[j] > max_val:
                max_val = high[j]
        if valid:
            hh[i] = max_val
        else:
            hh[i] = np.nan

    # Compute rolling minimum of low over 'period'
    ll = np.empty(n, dtype=np.float64)
    for i in range(n):
        ll[i] = np.nan
    for i in range(period - 1, n):
        min_val = 1e100  # very high number as initial min
        valid = True
        for j in range(i - period + 1, i + 1):
            if np.isnan(low[j]):
                valid = False
                break
            if low[j] < min_val:
                min_val = low[j]
        if valid:
            ll[i] = min_val
        else:
            ll[i] = np.nan

    # Compute the Choppiness Index result
    res = np.empty(n, dtype=np.float64)
    for i in range(n):
        res[i] = np.nan
    log_period = np.log10(period)
    for i in range(period - 1, n):
        if np.isnan(atr_sum[i]) or np.isnan(hh[i]) or np.isnan(ll[i]) or (hh[i] - ll[i]) <= 0:
            res[i] = np.nan
        else:
            res[i] = (scalar * (np.log10(atr_sum[i]) - np.log10(hh[i] - ll[i]))) / log_period

    return res



import numpy as np
from typing import Union
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


@njit(cache=True)
def _rsi(p: np.ndarray, period: int) -> np.ndarray:
    """
    Compute the Relative Strength Index using a loop and Wilder's smoothing.
    """
    n = len(p)
    rsi_arr = np.full(n, np.nan)
    if n < period + 1:
        return rsi_arr
    # Calculate differences between consecutive prices.
    diff = np.empty(n - 1)
    for i in range(n - 1):
        diff[i] = p[i+1] - p[i]

    # Compute initial average gain and loss over the first 'period' differences.
    sum_gain = 0.0
    sum_loss = 0.0
    for i in range(period):
        change = diff[i]
        if change > 0:
            sum_gain += change
        else:
            sum_loss += -change
    avg_gain = sum_gain / period
    avg_loss = sum_loss / period

    # Compute first RSI value at index 'period'
    if avg_loss == 0:
        rsi_arr[period] = 100.0
    else:
        rs = avg_gain / avg_loss
        rsi_arr[period] = 100 - (100 / (1 + rs))

    # Recursively update average gain and loss and compute subsequent RSI values.
    for i in range(period, n - 1):
        change = diff[i]
        gain = change if change > 0 else 0.0
        loss = -change if change < 0 else 0.0
        avg_gain = (avg_gain * (period - 1) + gain) / period
        avg_loss = (avg_loss * (period - 1) + loss) / period
        if avg_loss == 0:
            rsi_arr[i+1] = 100.0
        else:
            rs = avg_gain / avg_loss
            rsi_arr[i+1] = 100 - (100 / (1 + rs))
    return rsi_arr


def rsi(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    RSI - Relative Strength Index using Numba for optimization

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    p = np.asarray(source, dtype=float)
    result = _rsi(p, period)
    return result if sequential else result[-1]
